from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Generic, TypeAlias, TypeVar

TRes = TypeVar("TRes")
TRequest = TypeVar("TRequest")


HandleNext: TypeAlias = Callable[[TRequest], TRes]


class PipelineBehavior(Generic[TRequest, TRes], ABC):
    @abstractmethod
    def handle(
        self,
        request: TRequest,
        handle_next: HandleNext[TRequest, TRes],
    ) -> TRes: ...



from collections.abc import Reversible
from typing import Any

from bazario.abc.handler import NotificationHandler, RequestHandler
from bazario.abc.pipeline_behavior import (
    HandleNext,
    PipelineBehavior,
)


def build_pipeline_behaviors_chain(
    handler: RequestHandler | NotificationHandler,
    behaviors: Reversible[PipelineBehavior],
) -> HandleNext:
    current_behavior: HandleNext = handler.handle

    for behavior in reversed(behaviors):
        current_behavior = _wrap_with_behavior(behavior, current_behavior)

    return current_behavior


def _wrap_with_behavior(
    behavior: PipelineBehavior,
    handle_next: HandleNext,
) -> HandleNext:
    def wrapped_handler(request: Any) -> Any:
        return behavior.handle(request, handle_next)

    return wrapped_handler



from bazario.abc.handler import NotificationHandler, RequestHandler
from bazario.abc.pipeline_behavior import HandleNext, PipelineBehavior
from bazario.abc.publisher import Publisher
from bazario.abc.resolver import Resolver
from bazario.abc.sender import Sender
from bazario.dispatcher import Dispatcher
from bazario.markers import Notification, Request
from bazario.registry import Registry

__all__ = (
    "Dispatcher",
    "HandleNext",
    "Notification",
    "NotificationHandler",
    "PipelineBehavior",
    "Publisher",
    "Registry",
    "Request",
    "RequestHandler",
    "Resolver",
    "Sender",
)



from typing import TypeVar

from dishka import AsyncContainer

from bazario.asyncio.abc.resolver import Resolver

T = TypeVar("T")


class DishkaResolver(Resolver):
    def __init__(self, container: AsyncContainer) -> None:
        self._container = container

    async def resolve(self, dependency_type: type[T]) -> T:
        return await self._container.get(dependency_type)



from bazario.abc.handler import NotificationHandler, RequestHandler
from bazario.abc.pipeline_behavior import PipelineBehavior
from bazario.markers import Notification, Request


class Registry:
    request_handlers: dict[type[Request], type[RequestHandler]]
    notification_handlers: dict[
        type[Notification],
        list[type[NotificationHandler]],
    ]
    pipeline_behaviors: dict[type, list[type[PipelineBehavior]]]

    def __init__(self) -> None:
        self.request_handlers = {}
        self.notification_handlers = {}
        self.pipeline_behaviors = {}

    def add_request_handler(
        self,
        request_type: type[Request],
        request_handler: type[RequestHandler],
    ) -> None:
        self.request_handlers[request_type] = request_handler

    def add_notification_handlers(
        self,
        notification_type: type[Notification],
        *notification_handlers: type[NotificationHandler],
    ) -> None:
        self.notification_handlers.setdefault(notification_type, []).extend(
            notification_handlers,
        )

    def add_pipeline_behaviors(
        self,
        request_type: type,
        *pipeline_behaviors: type[PipelineBehavior],
    ) -> None:
        self.pipeline_behaviors.setdefault(request_type, []).extend(
            pipeline_behaviors,
        )

    def get_request_handler(
        self,
        request_type: type[Request],
    ) -> type[RequestHandler] | None:
        return self.request_handlers.get(request_type)

    def get_notification_handlers(
        self,
        notification_type: type[Notification],
    ) -> list[type[NotificationHandler]]:
        return self.notification_handlers.get(notification_type, [])

    def get_pipeline_behaviors(
        self,
        request_type: type,
    ) -> list[type[PipelineBehavior]]:
        return [
            behavior
            for request, behaviors in self.pipeline_behaviors.items()
            for behavior in behaviors
            if issubclass(request_type, request)
        ]



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from bazario.markers import Notification, Request

TRes_co = TypeVar("TRes_co", covariant=True)
TReq_contra = TypeVar("TReq_contra", bound=Request, contravariant=True)
TNot_contra = TypeVar("TNot_contra", bound=Notification, contravariant=True)


class RequestHandler(Generic[TReq_contra, TRes_co], ABC):
    @abstractmethod
    def handle(self, request: TReq_contra) -> TRes_co: ...


class NotificationHandler(Generic[TNot_contra], ABC):
    @abstractmethod
    def handle(self, notification: TNot_contra) -> None: ...



from bazario.markers import Notification, Request


class HandlerNotFoundError(Exception):
    def __init__(
        self,
        target_type: type[Request] | type[Notification],
    ) -> None:
        self.target_type = target_type

        super().__init__(
            f"Handler for target '{target_type.__name__}' not found.",
        )



from abc import ABC, abstractmethod
from typing import TypeVar

TDependency = TypeVar("TDependency")


class Resolver(ABC):
    @abstractmethod
    async def resolve(
        self,
        dependency_type: type[TDependency],
    ) -> TDependency: ...



from typing import TypeVar

from punq import Container

from bazario import Resolver

TDependency = TypeVar("TDependency")


class PunqResolver(Resolver):
    def __init__(self, container: Container) -> None:
        self._container = container

    def resolve(self, dependency_type: type[TDependency]) -> TDependency:
        return self._container.resolve(dependency_type)



from abc import ABC, abstractmethod
from collections.abc import Callable, Coroutine
from typing import Any, Generic, TypeAlias, TypeVar

TRes = TypeVar("TRes")
TRequest = TypeVar("TRequest")


HandleNext: TypeAlias = Callable[[TRequest], Coroutine[Any, Any, TRes]]


class PipelineBehavior(Generic[TRequest, TRes], ABC):
    @abstractmethod
    async def handle(
        self,
        request: TRequest,
        handle_next: HandleNext[TRequest, TRes],
    ) -> TRes: ...



from abc import ABC, abstractmethod
from typing import TypeVar

TDependency = TypeVar("TDependency")


class Resolver(ABC):
    @abstractmethod
    def resolve(self, dependency_type: type[TDependency]) -> TDependency: ...



from bazario.asyncio.abc.publisher import Publisher
from bazario.asyncio.abc.resolver import Resolver
from bazario.asyncio.abc.sender import Sender, TRes
from bazario.asyncio.chain import build_pipeline_behaviors_chain
from bazario.asyncio.registry import Registry
from bazario.exceptions import HandlerNotFoundError
from bazario.markers import Notification, Request


class Dispatcher(Sender, Publisher):
    def __init__(self, resolver: Resolver, registry: Registry) -> None:
        self._resolver = resolver
        self._registry = registry

    async def send(self, request: Request[TRes]) -> TRes:
        request_type = type(request)

        handler_class = self._registry.get_request_handler(request_type)

        if not handler_class:
            raise HandlerNotFoundError(request_type)

        handler = await self._resolver.resolve(handler_class)
        behaviors = [
            await self._resolver.resolve(behavior_type)
            for behavior_type in self._registry.get_pipeline_behaviors(
                request_type,
            )
        ]
        handle_next = build_pipeline_behaviors_chain(handler, behaviors)

        return await handle_next(request)

    async def publish(self, notification: Notification) -> None:
        notification_type = type(notification)

        handler_classes = self._registry.get_notification_handlers(
            notification_type,
        )
        for handler_class in handler_classes:
            handler = await self._resolver.resolve(handler_class)
            behaviors = [
                await self._resolver.resolve(behavior_type)
                for behavior_type in self._registry.get_pipeline_behaviors(
                    notification_type,
                )
            ]
            handle_next = build_pipeline_behaviors_chain(handler, behaviors)

            await handle_next(notification)
